from enum import StrEnum


class Currency(StrEnum):
    # none = "none"

    IRR = "IRR"
    # IRT = "IRT"

    USD = "USD"
    EUR = "EUR"
    # GBP = "GBP"

    # USDT = "USDT"
    # BTC = "BTC"
    # ETH = "ETH"

    @classmethod
    def main_currency(cls) -> "Currency":
        return cls.IRR

    @property
    def properties(self) -> dict[str, dict[str, str | int | bool]]:
        return {
            Currency.IRR: {
                "name": {
                    "fa": "ریال",
                    "en": "Iranian Rial",
                },
                "symbol": "IRR",
                "precision": 0,
                "icon": "https://flagcdn.com/w40/ir.png",
                "is_crypto": False,
                "color": "#1976d2",
            },
            Currency.USD: {
                "name": {
                    "fa": "دلار",
                    "en": "Dollar",
                },
                "symbol": "USD",
                "precision": 2,
                "icon": "https://flagcdn.com/w40/us.png",
                "is_crypto": False,
                "color": "#f7931a",
            },
            Currency.EUR: {
                "name": {
                    "fa": "یورو",
                    "en": "Euro",
                },
                "symbol": "EUR",
                "precision": 2,
                "icon": "https://flagcdn.com/w40/eu.png",
                "is_crypto": False,
                "color": "#26a17b",
            },
        }

    @property
    def currency(self) -> "Currency":
        return self

    @property
    def name(self) -> dict:
        return self.properties.get(self, {}).get("name")

    @property
    def symbol(self) -> str:
        return self.properties.get(self, {}).get("symbol")

    @property
    def precision(self) -> int:
        return self.properties.get(self, {}).get("precision")

    @property
    def icon(self) -> str:
        return self.properties.get(self, {}).get("icon")

    @property
    def is_crypto(self) -> bool:
        return self.properties.get(self, {}).get("is_crypto")

    @property
    def color(self) -> str:
        return self.properties.get(self, {}).get("color")


class StatusEnum(StrEnum):
    ACTIVE = "active"
    INACTIVE = "inactive"
    SUSPENDED = "suspended"



from datetime import datetime
from decimal import Decimal
from enum import StrEnum

from fastapi_mongo_base.schemas import TenantUserEntitySchema
from fastapi_mongo_base.utils import bsontools, timezone
from pydantic import BaseModel, field_validator

from .enums import Currency


class HoldStatus(StrEnum):
    ACTIVE = "active"
    EXPIRED = "expired"
    CANCELLED = "cancelled"
    RELEASED = "released"
    FAILED = "failed"


class WalletHoldSchema(TenantUserEntitySchema):
    wallet_id: str
    currency: str
    amount: Decimal
    expires_at: datetime | None = None
    status: HoldStatus = HoldStatus.ACTIVE
    description: str | None = None

    @field_validator("amount", mode="before")
    def validate_amount(cls, value: Decimal) -> Decimal:  # noqa: N805
        return bsontools.decimal_amount(value)

    def is_expired(self) -> bool:
        return (
            self.expires_at < datetime.now(timezone.tz)
            if self.expires_at
            else False
        )


class WalletHoldCreateSchema(BaseModel):
    currency: Currency
    amount: Decimal
    expires_at: datetime | None = None
    status: HoldStatus = HoldStatus.ACTIVE
    meta_data: dict | None = None
    description: str | None = None


class WalletHoldUpdateSchema(BaseModel):
    expires_at: datetime | None = None
    status: HoldStatus | None = None
    meta_data: dict | None = None
    description: str | None = None



from decimal import Decimal
from enum import StrEnum

from fastapi_mongo_base.schemas import TenantUserEntitySchema
from fastapi_mongo_base.tasks import TaskMixin
from fastapi_mongo_base.utils import bsontools
from pydantic import BaseModel, field_validator


class ProposalStatus(StrEnum):
    draft = "draft"
    init = "init"
    processing = "processing"
    completed = "completed"
    failed = "failed"
    error = "error"


class Participant(BaseModel):
    wallet_id: str
    amount: Decimal
    hold_id: str | None = None

    @field_validator("amount", mode="before")
    def validate_amount(cls, value: Decimal) -> Decimal:  # noqa: N805
        return bsontools.decimal_amount(value)


class ProposalSchema(TenantUserEntitySchema, TaskMixin):
    issuer_id: str
    amount: Decimal
    description: str | None = None
    note: str | None = None
    currency: str
    status: ProposalStatus = ProposalStatus.init
    participants: list[Participant]

    @field_validator("amount", mode="before")
    def validate_amount(cls, value: Decimal) -> Decimal:  # noqa: N805
        return bsontools.decimal_amount(value)


class ProposalCreateSchema(BaseModel):
    amount: Decimal
    description: str | None = None
    note: str | None = None
    currency: str
    status: ProposalStatus = ProposalStatus.init
    participants: list[Participant]
    meta_data: dict | None = None


class ProposalUpdateSchema(BaseModel):
    # status: str | None
    status: ProposalStatus | None = None
    description: str | None = None
    note: str | None = None
    meta_data: dict | None = None



from decimal import Decimal
from enum import StrEnum

from fastapi_mongo_base.schemas import TenantUserEntitySchema
from pydantic import BaseModel, Field, model_validator

from .enums import Currency


class WalletOwnerType(StrEnum):
    user = "user"
    tenant = "tenant"
    app = "app"
    system = "system"


class WalletPurpose(StrEnum):
    regular = "regular"
    treasury = "treasury"
    test = "test"

    # settlement = "settlement"
    # fee_collector = "fee_collector"
    # income_wallet = "income_wallet"
    # bonus_wallet = "bonus_wallet"
    # reserve = "reserve"


class WalletBalanceType(StrEnum):
    positive = "positive"
    negative = "negative"


class WalletStatus(StrEnum):
    active = "active"
    inactive = "inactive"
    pending = "pending"
    blocked = "blocked"
    deleted = "deleted"


class WalletSchema(TenantUserEntitySchema):
    wallet_purpose: WalletPurpose = WalletPurpose.regular
    owner_type: WalletOwnerType = WalletOwnerType.user
    balance_type: WalletBalanceType = WalletBalanceType.positive
    status: WalletStatus = WalletStatus.active

    main_currency: Currency = Currency.main_currency()
    is_default: bool = True


class BalanceSchema(BaseModel):
    currency: Currency
    total: Decimal = Field(ge=0)
    held: Decimal = Field(ge=0)
    available: Decimal = Field(ge=0)

    @model_validator(mode="before")
    def validate_balance(cls, values: dict[str, object]) -> dict[str, object]:  # noqa: N805
        total = Decimal(values.get("total"))
        held = Decimal(values.get("held"))
        available = Decimal(values.get("available"))
        set_fields = [x is not None for x in (total, held, available)]
        num_set = sum(set_fields)

        if num_set < 2:
            raise ValueError(
                "At least two of 'total', 'held', and 'available' must be set."
            )

        if num_set == 2:
            if total is None:
                values["total"] = held + available
            elif held is None:
                values["held"] = total - available
            elif available is None:
                values["available"] = total - held
        elif num_set == 3 and (total != held + available):
            raise ValueError(
                f"'total' ({total}) must equal "
                f"'held' ({held}) + 'available' ({available})"
            )
        return values


class WalletDetailSchema(WalletSchema):
    balance: dict[str, BalanceSchema] = Field(default_factory=dict)


class WalletCreateSchema(BaseModel):
    user_id: str | None = None
    meta_data: dict | None = None

    wallet_purpose: WalletPurpose = WalletPurpose.regular
    owner_type: WalletOwnerType = WalletOwnerType.user
    balance_type: WalletBalanceType = WalletBalanceType.positive
    status: WalletStatus = WalletStatus.active

    main_currency: Currency = Currency.main_currency()
    is_default: bool = False


class WalletUpdateSchema(BaseModel):
    meta_data: dict | None = None
    is_default: bool | None = None



import os
from datetime import datetime

import httpx
from usso.utils import agent

from .hold import HoldStatus, WalletHoldCreateSchema, WalletHoldSchema
from .proposal import Participant, ProposalCreateSchema, ProposalSchema
from .wallet import WalletDetailSchema


class AccountingClient(httpx.AsyncClient):
    def __init__(self, tenant_id: str) -> None:
        accounting_service_url = os.getenv(
            "ACCOUNTING_SERVICE_URL", "https://wallets.uln.me"
        )
        super().__init__(
            base_url=f"{accounting_service_url}/api/accounting/v1"
        )
        self.tenant_id = tenant_id

    async def get_token(self, scopes: str | list[str]) -> str:
        if isinstance(scopes, str):
            scopes = [scopes]

        jwt = agent.generate_agent_jwt(
            scopes=scopes,
            aud="accounting",
            tenant_id=self.tenant_id,
        )
        token = await agent.get_agent_token_async(jwt)
        self.headers["Authorization"] = f"Bearer {token}"
        return token

    async def get_wallet(self, wallet_id: str) -> WalletDetailSchema:
        await self.get_token("read:finance/accounting/wallet")
        response = await self.get(f"/wallets/{wallet_id}")
        response.raise_for_status()
        return WalletDetailSchema.model_validate(response.json())

    async def get_holds(self, wallet_id: str) -> list[WalletHoldSchema]:
        await self.get_token("read:finance/accounting/hold")
        response = await self.get(f"/wallets/{wallet_id}/holds")
        response.raise_for_status()
        return [
            WalletHoldSchema.model_validate(item)
            for item in response.json().get("items", [])
        ]

    async def total_held_amount(
        self,
        wallet_id: str,
        currency: str,
        status: HoldStatus = HoldStatus.ACTIVE,
    ) -> float:
        holds = await self.get_holds(wallet_id)
        return sum(
            hold.amount
            for hold in holds
            if hold.currency == currency and hold.status == status
        )

    async def create_hold(
        self,
        wallet_id: str,
        currency: str,
        amount: float,
        expires_at: datetime,
    ) -> WalletHoldSchema:
        await self.get_token("create:finance/accounting/hold")
        response = await self.post(
            f"/wallets/{wallet_id}/holds",
            json=WalletHoldCreateSchema(
                currency=currency,
                amount=amount,
                expires_at=expires_at,
                status=HoldStatus.ACTIVE,
            ).model_dump(mode="json"),
        )
        response.raise_for_status()
        return WalletHoldSchema.model_validate(response.json())

    async def create_proposal(
        self,
        *,
        from_wallet_id: str,
        to_wallet_id: str,
        currency: str,
        amount: float,
        description: str | None = None,
        note: str | None = None,
        hold_id: str | None = None,
    ) -> ProposalSchema:
        await self.get_token("create:finance/accounting/proposal")
        response = await self.post(
            "/proposals",
            json=ProposalCreateSchema(
                participants=[
                    Participant(
                        wallet_id=from_wallet_id,
                        amount=-amount,
                        hold_id=hold_id,
                    ),
                    Participant(wallet_id=to_wallet_id, amount=amount),
                ],
                amount=amount,
                currency=currency,
                description=description,
                note=note,
            ).model_dump(mode="json"),
        )
        response.raise_for_status()
        return ProposalSchema.model_validate(response.json())
